/*
 * Copyright 2018 The TensorFlow Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.tensorflow.demo;

import android.graphics.Bitmap;
import android.graphics.Bitmap.Config;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Paint.Style;
import android.graphics.RectF;
import android.graphics.Typeface;
import android.media.ImageReader.OnImageAvailableListener;
import android.os.SystemClock;
import android.util.Size;
import android.util.TypedValue;
import android.widget.ImageView;
import android.widget.Toast;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;
import org.tensorflow.demo.OverlayView.DrawCallback;
import org.tensorflow.demo.env.BorderedText;
import org.tensorflow.demo.env.ImageUtils;
import org.tensorflow.demo.env.Logger;
import org.tensorflow.demo.tracking.MultiBoxTracker;
import org.tensorflow.lite.demo.R; // Explicit import needed for internal Google builds.

/**
 * An activity that uses a TensorFlowMultiBoxDetector and ObjectTracker to detect and then track
 * objects.
 */
public class DetectorActivity extends CameraActivity implements OnImageAvailableListener {
  private static final Logger LOGGER = new Logger();

  // Configuration values for the prepackaged SSD model.
//  private static final int TF_OD_API_INPUT_SIZE = 300;
//  private static final boolean TF_OD_API_IS_QUANTIZED = true;
//  private static final String TF_OD_API_MODEL_FILE = "detect.tflite";
//  private static final String TF_OD_API_LABELS_FILE = "file:///android_asset/coco_labels_list.txt";

  // Configuration values for the prepackaged SSD model.
  private static final int TF_OD_API_INPUT_SIZE = 224;//300;
  private static final boolean TF_OD_API_IS_QUANTIZED = true;
  private static final String TF_OD_API_MODEL_FILE = "detect.tflite";
  private static final String TF_OD_API_LABELS_FILE = "file:///android_asset/coco_labels_list.txt";


  // Which detection model to use: by default uses Tensorflow Object Detection API frozen
  // checkpoints.
  private enum DetectorMode {
    TF_OD_API;
  }

  private static final DetectorMode MODE = DetectorMode.TF_OD_API;

  // Minimum detection confidence to track a detection.
  private static final float MINIMUM_CONFIDENCE_TF_OD_API = 0.6f;

  private static final boolean MAINTAIN_ASPECT = false;

  private static final Size DESIRED_PREVIEW_SIZE = new Size(640, 480);
//private static final Size DESIRED_PREVIEW_SIZE = new Size(1080, 720);

  private static final boolean SAVE_PREVIEW_BITMAP = false;
  private static final float TEXT_SIZE_DIP = 10;

  private Integer sensorOrientation;



  private long lastProcessingTimeMs;
  private Bitmap rgbFrameBitmap = null;
  private Bitmap croppedBitmap = null;
  private Bitmap cropCopyBitmap = null;
  private Bitmap convertBitmap = null;
  private boolean computingDetection = false;

  private long timestamp = 0;

  private Matrix frameToCropTransform;
  private Matrix cropToFrameTransform;

  private MultiBoxTracker tracker;
  private Classifier detector;///////////////////////////////

  private byte[] luminanceCopy;

  private BorderedText borderedText;
  @Override
  public void onPreviewSizeChosen(final Size size, final int rotation) {
    LOGGER.i("rotation : %d",rotation);

    final float textSizePx =
        TypedValue.applyDimension(//这个方法的作用是 把Android系统中的非标准度量尺寸转变为标准度量尺寸 (Android系统中的标准尺寸是px, 即像素)
            TypedValue.COMPLEX_UNIT_DIP, TEXT_SIZE_DIP, getResources().getDisplayMetrics());
    borderedText = new BorderedText(textSizePx);
    borderedText.setTypeface(Typeface.MONOSPACE);

    tracker = new MultiBoxTracker(this);

    int cropSize = TF_OD_API_INPUT_SIZE;

//    try {
      detector =
          TFLiteObjectDetectionAPIModel.create(
              getAssets(),
              TF_OD_API_MODEL_FILE,
              TF_OD_API_LABELS_FILE,
              TF_OD_API_INPUT_SIZE,
              TF_OD_API_IS_QUANTIZED);
      cropSize = TF_OD_API_INPUT_SIZE;
//    } catch (final IOException e) {
//      LOGGER.e("Exception initializing classifier!", e);
//      Toast toast = Toast.makeText(getApplicationContext(), "Classifier could not be initialized", Toast.LENGTH_SHORT);
//      toast.show();
//      finish();
//    }


    previewWidth = size.getWidth();
    previewHeight = size.getHeight();

    sensorOrientation = rotation - getScreenOrientation();
   // LOGGER.i("Camera orientation relative to screen canvas: %d", sensorOrientation);

   // LOGGER.i("Initializing at size %dx%d", previewWidth, previewHeight);
    rgbFrameBitmap = Bitmap.createBitmap(previewWidth, previewHeight, Config.ARGB_8888);
    croppedBitmap = Bitmap.createBitmap(cropSize, cropSize, Config.ARGB_8888);
    //convertBitmap = Bitmap.createBitmap(cropSize, cropSize, Config.ARGB_8888);
    frameToCropTransform =
        ImageUtils.getTransformationMatrix(
            previewWidth, previewHeight,
            cropSize, cropSize,
            sensorOrientation, MAINTAIN_ASPECT);

    cropToFrameTransform = new Matrix();
    frameToCropTransform.invert(cropToFrameTransform);

    trackingOverlay = (OverlayView) findViewById(R.id.tracking_overlay);
    trackingOverlay.addCallback(
        new DrawCallback() {
          @Override
          public void drawCallback(final Canvas canvas) {
            tracker.draw(canvas);
//            if (isDebug()) {////////////////////
//              tracker.drawDebug(canvas);
//            }
          }
        });

//    addCallback(
//        new DrawCallback() {
//          @Override
//          public void drawCallback(final Canvas canvas) {
//            LOGGER.i("######drawCallback###### :%s",isDebug());
//            if (!isDebug()) {
//              LOGGER.i("Frame is %d ms", lastProcessingTimeMs);
//              return;
//            }
//            //-----------这里画出 进行识别和跟踪的小图片，实际中，其是隐藏的
//            final Bitmap copy = cropCopyBitmap;
//            if (copy == null) {
//              return;
//            }
//
//            final int backgroundColor = Color.argb(100, 0, 0, 0);
//            canvas.drawColor(backgroundColor);
//
//            final Matrix matrix = new Matrix();
//            final float scaleFactor = 2;
//            matrix.postScale(scaleFactor, scaleFactor);
//            matrix.postTranslate(
//                canvas.getWidth() - copy.getWidth() * scaleFactor,
//                canvas.getHeight() - copy.getHeight() * scaleFactor);
//            canvas.drawBitmap(copy, matrix, new Paint());
//
//            final Vector<String> lines = new Vector<String>();
//            if (detector != null) {
//              final String statString = detector.getStatString();
//              final String[] statLines = statString.split("\n");
//              for (final String line : statLines) {
//                lines.add(line);
//              }
//            }
//            lines.add("");
//
//            lines.add("Frame: " + previewWidth + "x" + previewHeight);
//            lines.add("Crop: " + copy.getWidth() + "x" + copy.getHeight());
//            lines.add("View: " + canvas.getWidth() + "x" + canvas.getHeight());
//            lines.add("Rotation: " + sensorOrientation);
//            lines.add("Inference time: " + lastProcessingTimeMs + "ms");
//
//            borderedText.drawLines(canvas, 10, canvas.getHeight() - 10, lines);///////////QQQQQQQQQQQQQQQQQQQ
//
//          }
//
//        });
  }

  OverlayView trackingOverlay;

  @Override
  protected void processImage() {
    ++timestamp;
    final long currTimestamp = timestamp;
    byte[] originalLuminance = getLuminance();//即yuvBytes[0]
    LOGGER.i("processImage: %d",getLuminanceStride());
    tracker.onFrame(
        previewWidth,
        previewHeight,
        getLuminanceStride(),
        sensorOrientation,
        originalLuminance,
        timestamp);
    //// 使用postInvalidate可以直接在线程中更新界面
    trackingOverlay.postInvalidate();

    // No mutex needed as this method is not reentrant.
    if (computingDetection) {//如果正在处理检测，则此处就只进行preview了！
      readyForNextImage();
      return;
    }
    computingDetection = true;
    LOGGER.i("Preparing image " + currTimestamp + " for detection in bg thread.");

    rgbFrameBitmap.setPixels(getRgbBytes(), 0, previewWidth, 0, 0, previewWidth, previewHeight);

    if (luminanceCopy == null) {
      luminanceCopy = new byte[originalLuminance.length];
    }
    //public static void arraycopy(Object src,  int srcPos,Object dest,int destPos,int length)
    System.arraycopy(originalLuminance, 0, luminanceCopy, 0, originalLuminance.length);//这里进行图像拷贝，为跟踪做准备
    readyForNextImage();//进行Camera 的preview

    final Canvas canvas = new Canvas(croppedBitmap);
    canvas.drawBitmap(rgbFrameBitmap, frameToCropTransform, null);
    // For examining the actual TF input.
//    if (SAVE_PREVIEW_BITMAP) {///-
//      ImageUtils.saveBitmap(croppedBitmap);
//    }

    runInBackground(//这部分是在线程中运行的！！！
        new Runnable() {
          @Override
          public void run() {
            LOGGER.i("Running detection on image " + currTimestamp);

            final long startTime = SystemClock.uptimeMillis();
            final List<Classifier.Recognition> results = detector.recognizeImage(croppedBitmap);
            lastProcessingTimeMs = SystemClock.uptimeMillis() - startTime;

            cropCopyBitmap = Bitmap.createBitmap(croppedBitmap);
//            Canvas canvas=null;
//            Paint paint=null;
//            if(isDebug()) {
//            canvas = new Canvas(cropCopyBitmap);
//            paint = new Paint();
//            paint.setColor(Color.RED);
//            paint.setStyle(Style.STROKE);
//            paint.setStrokeWidth(6.0f);
//            }
            float minimumConfidence = MINIMUM_CONFIDENCE_TF_OD_API;
            switch (MODE) {
              case TF_OD_API:
                minimumConfidence = MINIMUM_CONFIDENCE_TF_OD_API;
                break;
            }

            final List<Classifier.Recognition> mappedRecognitions = new LinkedList<Classifier.Recognition>();

            for (final Classifier.Recognition result : results) {
              final RectF location = result.getLocation();

              if (location != null && result.getConfidence() >= minimumConfidence) {
//                if (isDebug()) {
//                canvas.drawRect(location, paint);///////////////////////////////////////////这是画在小的图片上的Debug图片/////
//                LOGGER.i("image RectF %f,%f,%f,%f" , location.top , location.left , location.bottom , location.right);
//                LOGGER.i("mappedRecognitions:%f ,%s , %s" , result.getConfidence(),result.getId(),result.getTitle());
//                }
                cropToFrameTransform.mapRect(location);
                result.setLocation(location);
                mappedRecognitions.add(result);
              }
            }
            //把位置信息给跟踪程序，进行跟踪
            tracker.trackResults(mappedRecognitions, luminanceCopy, currTimestamp);
            trackingOverlay.postInvalidate();

            requestRender();
            computingDetection = false;
          }
        });
  }

  @Override
  protected int getLayoutId() {
    return R.layout.camera_connection_fragment_tracking;
  }

  @Override
  protected Size getDesiredPreviewFrameSize() {
    return DESIRED_PREVIEW_SIZE;
  }

  @Override
  public void onSetDebug(final boolean debug) {
//    LOGGER.i("onSetDebug :"+debug);
//    detector.enableStatLogging(debug);
  }
}
